"""
Created by adam on 2/17/20
"""
from abc import ABC

import pandas as pd

from CanvasHacks.Repositories.mixins import TimeHandlerMixin
from CanvasHacks.TimeTools import check_is_date
from CanvasHacks.UtilityDecorators import ensure_timestamps

__author__ = 'adam'

if __name__ == '__main__':
    pass


class IPenalizer(ABC):


    def analyze( self, *args, **kwargs ):
        """
        THIS IS THE LATEST VERSION CAN-60
        Returns a signed float of the percentage
        the contribution the evaluated item makes to the
        overall score should be lowered.

        :param work:
        :return: signed float
        """
        raise NotImplementedError


    def get_penalty( self, submitted_date ):
        """

        Returns the percentage that the assignment
        should be penalized.
        Not used on its own. get_fudge_points and
        get_penalized_score properly apply the output of this
        method for the particular grading use.

        :type submitted_date: pd.Datetime
        """
        raise NotImplementedError

    def get_fudge_points( self, submitted_date, total_score, row=None ):
        """
        For assignment types where we have to submit the
        grade via fudge points, this calculates the amount for
        canvas to subtract or add to the total score.

        If row is provided, will save record to self.penalized_records

        :param submitted_date: The date the student work was submitted
        :param total_score: The score that we will apply penalty to
        :param row: None or pd.DataFrame row. If not None, will save record
        :return: float
        """
        raise NotImplementedError

    @property
    def penalty_messages( self ):
        """
        Returns a list of penalty messages
        :return: list
        """
        out = []

        if len( self.penalized_records ) > 0:
            for penalty_dict in self.penalized_records:
                due_at = None
                try:
                    due_at = self.due_date
                except AttributeError:
                    # Some penalizers may not have a due date
                    pass
                msg = self._make_penalty_message( penalty_dict[ 'penalty' ], penalty_dict[ 'record' ], due_at=None )
                out.append(msg)
        return out


    def _make_penalty_message( self, penalty, row, due_at=None ):
            """
            Handles printing or logging of penalties applied

            # todo enable logging

            :param penalty:
            :param row:
            :return:
            """
            stem = 'Student #{}: Submitted on {}; was due {}. Penalized {}'
            return stem.format( row[ 'student_id' ], row[ 'submitted' ], due_at, penalty )


class NoLatePenalty( IPenalizer ):
    """
    All graders will need a penalizer object
    so this can be used if there is no penalty for
    late submissions
    """

    def __init__( self, *args, **kwargs ):
        self.penalized_records = [ ]

    def get_penalty( self, submitted_date ):
        return 0

    def get_penalized_score( self, submitted_date, original_score ):
        return original_score

    def get_fudge_points( self, submitted_date, total_score, row=None ):
        """
        For assignment types where we have to submit the
        grade via fudge points, this calculates the amount for
        canvas to subtract or add to the total score.

        If row is provided, will save record to self.penalized_records

        :param submitted_date: The date the student work was submitted
        :param total_score: The score that we will apply penalty to
        :param row: None or pd.DataFrame row. If not None, will save record
        :return: float
        """
        return 0


class HalfLate( IPenalizer, TimeHandlerMixin ):
    """Late assignments receive half credit"""

    def __init__( self, due_date, grace_period=None ):
        due_date = self._force_timestamp( due_date )
        assert (isinstance( due_date, pd.Timestamp ))
        self.due_date = due_date

        if grace_period is not None:
            assert (isinstance( grace_period, pd.Timedelta ))
            self.due_date += grace_period

        # Record of applied penalties
        self.penalized_records = [
            # List of dicts with format
            # 'record': row,
            # 'penalty': penalty,
            # optional (for methods which used fudge points)
            # 'fudge_points': fudge_points
        ]

    def get_penalty( self, submitted_at ):
        """
        Returns the percentage that the assignment
        should be penalized.
        Not used on its own. get_fudge_points and
        get_penalized_score properly apply the output of this
        method for the particular grading use.

        Will return 0.0 for an on-time assignment
        :param submitted_at:
        :return:
        """
        submitted_at = self._force_timestamp( submitted_at )
        assert (isinstance( submitted_at, pd.Timestamp ))
        # Check if full credit
        if submitted_at <= self.due_date:
            return 0
        return 0.50

    def get_penalized_score( self, submitted_date, original_score, record=None ):
        """
        Returns the score reduced by the appropriate penalty.
        If record is not none, the penalty etc will be saved to
        penalized_records
        :param submitted_date:
        :param original_score:
        :param record:
        :return: float
        """
        penalty = self.get_penalty( submitted_date )
        # penalty was set up for uploading where have to use fudge points.
        # so we need to interpret it a bit here.
        # It will have returned 0 for no penalty and .5 for half credit
        if penalty == 0:
            return original_score
        elif penalty > 0 and record is not None:
            self.penalized_records.append(
                { 'record': record,
                  'penalty': penalty
                  } )
        return original_score * penalty


    def get_point_reduction_pct( self, submitted_date ):
        return 1 - self.get_penalty(submitted_date)

    def get_fudge_points( self, submitted_date, total_score, row=None ):
        """
        For assignment types where we have to submit the
        grade via fudge points, this calculates the amount for
        canvas to subtract or add to the total score.

        If row is provided, will save record to self.penalized_records

        :param submitted_date: The date the student work was submitted
        :param total_score: The score that we will apply penalty to
        :param row: None or pd.DataFrame row. If not None, will save record
        :return: float
        """
        # compute penalty if needed
        penalty = self.get_penalty( submitted_date )

        # will be 0 if not docking for lateness
        fudge_points = total_score * -penalty

        if penalty > 0 and row is not None:
            # Save record so calling class can handle
            self.penalized_records.append( {
                'record': row,
                'penalty': penalty,
                'fudge_points': fudge_points
            } )

        return fudge_points


class QuarterLate( IPenalizer, TimeHandlerMixin ):
    """
    Gives half credit for submissions between
    due date and last half credit date.
    Gives quarter credit thereafter
    """

    def __init__( self, due_date, last_half_credit_date, grace_period=None ):
        # assert (isinstance( due_date, pd.Timestamp ))
        # assert (isinstance( last_half_credit_date, pd.Timestamp ))
        self.last_half_credit_date = self._force_timestamp( last_half_credit_date )
        self.due_date = self._force_timestamp( due_date )

        if grace_period is not None:
            assert (isinstance( grace_period, pd.Timedelta ))
            self.due_date += grace_period
            self.last_half_credit_date += grace_period

        # Record of applied penalties
        self.penalized_records = [
            # List of dicts with format
            # 'record': row,
            # 'penalty': penalty,
            # optional if use fudge points
            # 'fudge_points': fudge_points
        ]


    def get_point_reduction_pct( self, submitted_date ):
        return 1 - self.get_penalty(submitted_date)

    def get_penalty( self, submitted_date ):
        """
        Computes the penalty as a percentage to be combined with
        the original score by another method.
        Not used on its own. get_fudge_points and
        get_penalized_score properly interpret the output of this
        method for the particular grading use.
        :param submitted_date:
        :return: float
        """
        submitted_date = self._force_timestamp( submitted_date )
        assert (isinstance( submitted_date, pd.Timestamp ))
        # Check if full credit
        if submitted_date <= self.due_date:
            return 0
        # if it was submitted after due date but before the quarter credit date
        # i.e, the first exam, it gets half
        if submitted_date <= self.last_half_credit_date:
            return 0.5
        # if it was after the half credit date, it gets quarter credit
        return 0.25

    def get_fudge_points( self, submitted_date, total_score, row=None ):
        """
        For assignment types where we have to submit the
        grade via fudge points, this calculates the amount for
        canvas to subtract or add to the total score.

        If row is provided, will save record to self.penalized_records

        :param submitted_date: The date the student work was submitted
        :param total_score: The score that we will apply penalty to
        :param row: None or pd.DataFrame row. If not None, will save record
        :return: float
        """
        # compute penalty if needed
        penalty = self.get_penalty( submitted_date )

        # penalty = self.penalizer.get_penalty(row['submitted'])
        # penalty = get_penalty( row[ 'submitted' ], self.activity.due_at, self.activity.last_half_credit_date, self.activity.grace_period )

        # will be 0 if not docking for lateness
        fudge_points = total_score * -penalty

        if penalty > 0 and row is not None:
            # Save record so calling class can handle
            self.penalized_records.append( {
                'record': row,
                'penalty': penalty,
                'fudge_points': fudge_points
            } )

        return fudge_points


# ================================= OLD


@ensure_timestamps
def get_penalty( submitted_date, due_date, last_half_credit_date, grace_period=None ):
    """
    OLD

    last_half_credit_date: Last moment they could receive half credit. If the unit hasn't closed, submissions after this will be given 0.25 credit. This can just be set to the lock date if there's no need to give quarter credit
    grace_period: Should be a pd.Timedelta object, e.g., pd.Timedelta('1 day').
    """
    assert (isinstance( submitted_date, pd.Timestamp ))

    if grace_period is not None:
        assert (isinstance( grace_period, pd.Timedelta ))
        due_date += grace_period
        last_half_credit_date += grace_period
    # Check if full credit
    if submitted_date <= due_date:
        return 0
    # if it was submitted after due date but before the quarter credit date
    # i.e, the first exam, it gets half
    if submitted_date <= last_half_credit_date:
        return .5
    # if it was after the half credit date, it gets quarter credit
    return .25



"""
Created by adam on 2/26/20
"""
__author__ = 'adam'

import pandas as pd
from canvasapi.quiz import QuizSubmission
from canvasapi.submission import Submission

if __name__ == '__main__':
    pass


class FrameStorageMixin:
    """
    Tools for use by repositories for which
    self.data is a dataframe of student records
    """

    def remove_student_records( self, student_ids ):
        student_ids = student_ids if isinstance( student_ids, list ) else [ student_ids ]
        if len( student_ids ) == 0:
            # bail if there were no students. That way we can run
            # this on first time
            return True

        is_index = False
        prelen = len( self.data )
        # figure out whether student_id is the index
        if 'student_id' not in self.data.columns:
            self.data.reset_index( inplace=True )
            is_index = True
        self.data = self.data[ ~self.data.student_id.isin( student_ids ) ].copy( deep=True )

        removed = prelen - len( self.data )
        if removed > 0:
            print( "Removed {} records".format( removed ) )

        if is_index:
            self.data.set_index( 'student_id', inplace=True )


class ObjectHandlerMixin:
    """Tools for dealing with
    input values which may represent students
    and other objects in a number of ways
    """

    def _handle_id( self, object_or_int ):
        """
        Takes either a object or the int value of their id
        and returns the id
        :param object_or_int:
        :return: int
        """
        try:
            return int( object_or_int )
        except TypeError:
            try:
                # in case we have a student object w id stored like this
                return object_or_int.student_id
            except AttributeError:
                return object_or_int.id

    def force_to_ids( self, list_of_students ):
        """We could receive a list of ids, Student
        objects or canvasapi User objects. This returns
        a list of ids"""
        out = [ ]

        for s in list_of_students:
            if isinstance( s, int ):
                out.append( s )
            else:
                # Both Student and User objects will have an id attribute
                # which contains the canvas id of the student
                out.append( s.id )
        return out


class TimeHandlerMixin:
    """Tools for classes which need to
    handle datetimes
    """

    def _force_timestamp( self, val ):
        """
        Forces the val to be a pandas timestamp
        :param val: string or timestamp
        :return: pd.Timestamp
        """
        if not isinstance( val, pd.Timestamp ):
            return pd.to_datetime( val )
        return val

    def _force_timedelta( self, val ):
        """
        Forces the provided value to be a timedelta
        :param val: string or pd.Timedelta
        :return: pd.Timedelta
        """
        if not isinstance( val, pd.Timedelta ):
            return pd.to_timedelta( val )


class StudentWorkMixin( ObjectHandlerMixin ):
    """Parent class for any repository which holds
    student data and can provide a formatted version
    for sending via email etc
    """

    @property
    def student_ids( self ):
        if isinstance( self.data, dict ):
            uids = [ k for k in self.data.keys() ]
        if isinstance( self.data, list ):
            uids = [ k[ 'student_id' ] for k in self.data ]
        if isinstance( self.data, pd.DataFrame ):
            try:
                uids = self.data.student_id.tolist()
            except KeyError:
                uids = self.data.reset_index()[ 'student_id' ].tolist()
        uids = list( set( uids ) )
        uids.sort()
        return uids

    def _check_empty( self, work ):
        """Checks whether the work is empty and
        returns the appropriate text to use
        """
        # Handle empty
        if work is None:
            return "THIS STUDENT SUBMISSION WAS BLANK. PLEASE GRADE ACCORDINGLY"
        return work

    # def store_results( self, results_list ):
    #     raise NotImplementedError


class SelectableMixin:
    """Allows to store a list of column names
    that have been specially designated by the user
    """

    def _init_selected( self ):
        try:
            if len( self.selected ) > 0:
                pass
        except Exception as e:
            print( e )
            self.selected = [ ]

    def select( self, identifier, name=None ):
        self._init_selected()
        self.selected.append( identifier )

    def deselect( self, identifier ):
        self.selected.pop( self.selected.index( identifier ) )

    def get_selections( self ):
        self._init_selected()
        return self.selected

    def reset_selections( self ):
        self.selected = [ ]




"""
Created by adam on 1/29/19
"""
__author__ = 'adam'

from CanvasHacks.Errors.grading import NonStringInContentField
from CanvasHacks.GradingAnalyzers.nonempty import NonEmpty
from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.Text.process import make_wordbag
from nltk.corpus import stopwords
import string
from CanvasHacks.GradingCorrections.penalities import get_penalty
from CanvasHacks.Text.stats import WordCount


class CreditForNonEmptyOLD( IGradingMethod ):
    """Returns full credit for a literally non-empty answer"""

    def __init__(self, min_words=2, count_stopwords=True):
        # self.score = 100
        # params
        self.min_words = min_words
        self.count_stopwords = count_stopwords

        # The object which will handle the actual processing and computation
        self.analyzer = NonEmpty
        # self.analyzer = WordCount(count_stopwords=count_stopwords)

    def grade( self, content, **kwargs ):
        """Returns the pct credit as an integer"""
        # Aliasing the pct int to leave room for
        # alternative output could go here

        return self.grade_pct_int(content, **kwargs)

    def grade_pct_int( self, content, on_credit=100, on_no_credit=None ):
        """Returns the pct credit as an integer
        Returns None if not getting any credit, which can be overridden
        by the value in on_no_credit
        """
        return on_credit if self.analyzer.analyze(content) else on_no_credit

        # if content is not None:
        #     if isinstance(content, (int, float)):
        #         return on_credit
        #
        #     if len(content) >0:
        #         return on_credit
        #
        # return on_no_credit

        # too processor intensive and duplicative
        # word_count = self.analyzer.analyze(content)
        # credit = word_count >= self.min_words
        #
        # # credit = receives_credit(content, self.min_words, self.count_stopwords)
        # if credit:
        #     return on_credit
        # elif on_no_credit is not None:
        #     return on_no_credit




class CreditForNonEmpty( IGradingMethod ):
    """Returns full credit for a literally non-empty answer"""

    def __init__(self, pct_of_score=1):
        # self.score = 100
        # params
        self.pct_of_score = pct_of_score

        # The object which will handle the actual processing and computation
        self.analyzer = NonEmpty
        # self.analyzer = WordCount(count_stopwords=count_stopwords)

    def grade( self, content, **kwargs ):
        """Returns the pct creditr"""
        return self.pct_of_score if self.analyzer.analyze(content) else 0

        # Aliasing the pct int to leave room for
        # alternative output could go here

        # return self.grade_pct_int(content, **kwargs)

    def grade_pct_int( self, content, on_credit=100, on_no_credit=None ):
        """Returns the pct credit as an integer
        Returns None if not getting any credit, which can be overridden
        by the value in on_no_credit
        """
        return on_credit if self.analyzer.analyze(content) else on_no_credit


# --------------------------- old

def receives_credit( content: str, min_words=2, count_stopwords=True ):
    """
    Given a piece of text written by a student, determines
    whether to assign credit or no credit

    :param content: The text to analyze
    :param min_words: The minimum word count to give credit for
    :param count_stopwords: Whether stopwords should count toward word count
    :return: Boolean
    """
    if not isinstance(content, str):
        raise NonStringInContentField
    remove = [ '``', "''", "'s"]
    remove += string.punctuation
    if not count_stopwords:
        remove += stopwords.words('english')
    remove = set(remove)
    bag = make_wordbag(content, remove)
    return len(bag) >= min_words


def scored_non_empty(content, max_score, on_empty=None):
    """
    Returns a dictionary with key 'score'
    :param content:
    :param max_score:
    :param on_empty: If none, will return no value if empty.
    :return:
    """
    if receives_credit( content ):
        return { 'score': max_score }
    else:
        if on_empty is not None:
            { 'score': on_empty }

        # questions[ qid ] = { 'score': 4.0 }
        # total_score += 4

def determine_credit(submissions):
    """
    OLD
    Adds submissions of zero length to the no-credit list.
    Adds others to the credit list.
    Returns a dictionary with keys credit and nocredit, with the lists as values
    """
    credit = []
    nocredit = []

    for s in submissions:
        if 'body' in s.keys() and s['body'] is not None and len(s['body']) > 2:
            credit.append(s['student_id'])
        else:
            nocredit.append(s['student_id'])

    return {'credit': credit, 'nocredit': nocredit}


def new_determine_journal_credit(activity, submissionRepo):
    """
    OLD
    Determines how much credit potentially late credit/no credit
    assignments should recieve.
    Created in CAN-24; moved into JournalGrader in CAN-40
    """
    results = []
    for submission in submissionRepo.data:
        if submission.body is not None:
            credit = receives_credit( submission.body )
            if credit:
                score = 100
            # Now check whether need to penalize for lateness
            penalty = get_penalty(submission.submitted_at, activity.due_at, activity.lock_at, activity.grace_period)
            # penalty was set up for uploading where have to use fudge points.
            # so we need to interpret it a bit here.
            # It will have returned 0 for no penalty and .5 for half credit
            penalty = 100 * penalty
            score = score - penalty
            results.append( (submission, int(score) ) )
    return results



if __name__ == '__main__':
    pass